from textual.app import ComposeResult
from textual.containers import VerticalScroll

from mastermind.game import Game
from mastermind.widgets.row import Row


class Board(VerticalScroll):
    def __init__(self, game: Game) -> None:
        super().__init__()

        self.game = game

        self.current_row_number = 1
        self.current_row = Row(self.game, row_number=self.current_row_number)

    def compose(self) -> ComposeResult:
        yield self.current_row

    def add_row(self):
        self.current_row_number += 1
        self.current_row = Row(self.game, row_number=self.current_row_number)
        self.mount(self.current_row)



import dataclasses

from textual.binding import Binding
from textual.widgets import Select
from textual.widgets._select import SelectOverlay

from mastermind.constants import BLANK_COLOR, CODE_PEG_COLORS
from mastermind.game import Game


class CodePeg(Select[int]):
    def __init__(self, game: Game) -> None:
        self.game = game

        super().__init__(
            options=zip(CODE_PEG_COLORS, range(1, self.game.num_colors + 1)),
            prompt=BLANK_COLOR,
            classes="code_peg",
        )

    def on_mount(self) -> None:
        option_list = self.query_one(SelectOverlay)

        esc_binding: Binding = option_list._bindings.key_to_bindings["escape"][0]
        option_list._bindings.key_to_bindings["escape"] = [
            dataclasses.replace(esc_binding, show=False)
        ]



from textual.widgets import Label

from mastermind.constants import CHECK_DEFAULT_TEXT, CHECK_HOVER_TEXT
from mastermind.game import Game


class Check(Label):
    def __init__(self, game: Game) -> None:
        self.game = game
        self.default_text = (f"{CHECK_DEFAULT_TEXT} " * self.game.num_pegs)[:-1]
        self.hover_text = (f"{CHECK_HOVER_TEXT} " * self.game.num_pegs)[:-1]

        super().__init__(self.default_text, id="check", classes="check")

    def on_enter(self) -> None:
        self.update(self.hover_text)

    def on_leave(self) -> None:
        self.update(self.default_text)



from pathlib import Path
from typing import Final

from textual.binding import Binding
from textual_utils import AppMetadata

APP_METADATA = AppMetadata(
    name="Master Mind",
    version="1.0.9",
    codename="🔴 ⚪",
    author="Rafal Padkowski",
    email="rafaelp@poczta.onet.pl",
)

ICON: Final[str] = "❔"

LOCALEDIR = Path(__file__).parent / "locale"

SETTINGS_PATH: Final[Path] = Path(__file__).parent / "settings" / "settings.toml"

BLANK_COLOR: Final[str] = "⭕"
CODE_PEG_COLORS: Final[list[str]] = ["🔴", "🟢", "🔵", "🟡", "🟣", "🟤", "⚪", "🟠"]
FEEDBACK_PEG_COLORS: Final[list[str]] = ["🔴", "⚪"]

CHECK_DEFAULT_TEXT: Final[str] = "❔"
CHECK_HOVER_TEXT: Final[str] = "❓"

KEY_TO_BINDING: Final[dict[str, Binding]] = {
    "ctrl+q": Binding(
        key="ctrl+q",
        action="quit",
        description="Quit",
        key_display="Ctrl+Q",
        show=True,
    ),
    "ctrl+c": Binding(
        key="ctrl+c",
        action="nothing",
        description="",
    ),
    "f2": Binding(
        key="f2",
        action="new_game",
        description="New game",
        key_display="F2",
    ),
    "f3": Binding(
        key="f3",
        action="settings",
        description="Settings",
        key_display="F3",
    ),
}



from dataclasses import dataclass, fields
from typing import Any, Final

from textual_utils import _

LANGUAGES: Final[dict[str, str]] = {
    "en": "English",
    "pl": "Polish",
}


@dataclass(frozen=True)
class Variation:
    name: str
    num_rows: int
    num_pegs: int
    num_colors: int

    @property
    def description(self) -> str:
        num_pegs_str = f"{self.num_pegs} pegs"

        return (
            f"{self.name} ({self.num_rows} {_('rows')}, "
            f"{_(num_pegs_str)}, {self.num_colors} {_('colors')})"
        )


VARIATIONS: Final[dict[str, Variation]] = {
    "original": Variation(name="original", num_rows=10, num_pegs=4, num_colors=6),
    "mini": Variation(name="mini", num_rows=6, num_pegs=4, num_colors=6),
    "super": Variation(name="super", num_rows=12, num_pegs=5, num_colors=8),
}


@dataclass
class Settings:
    language: str = "en"
    variation: Variation = VARIATIONS["original"]
    duplicate_colors: bool = False
    blank_color: bool = False

    def set(self, settings_dict: dict[str, Any]) -> None:
        for field in fields(self):
            if field.name == "variation":
                setattr(self, field.name, VARIATIONS[settings_dict[field.name]])
                continue

            setattr(self, field.name, settings_dict[field.name])


app_settings: Settings = Settings()



import dataclasses
from typing import Any

from textual import work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.events import Click
from textual.widget import Widget
from textual.widgets import Footer, Header, Label, Select, Switch
from textual_utils import (
    AboutHeaderIcon,
    ConfirmScreen,
    SettingRow,
    SettingsScreen,
    _,
    init_translation,
    load_settings,
    mount_about_header_icon,
    save_settings,
    set_translation,
)

from mastermind.constants import (
    APP_METADATA,
    BLANK_COLOR,
    CODE_PEG_COLORS,
    FEEDBACK_PEG_COLORS,
    ICON,
    KEY_TO_BINDING,
    LOCALEDIR,
    SETTINGS_PATH,
)
from mastermind.game import Game
from mastermind.settings import LANGUAGES, VARIATIONS, app_settings
from mastermind.widgets.board import Board


class MastermindApp(App):
    TITLE = "Master Mind"
    CSS_PATH = "styles.tcss"
    ENABLE_COMMAND_PALETTE = False
    BINDINGS = list(KEY_TO_BINDING.values())

    def __init__(self) -> None:
        super().__init__()

        settings_dict: dict[str, Any] = load_settings(SETTINGS_PATH)
        app_settings.set(settings_dict)

        init_translation(LOCALEDIR)
        set_translation(app_settings.language)
        self.translate_bindings()

        self.board: Board
        self.game: Game

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()

    async def on_mount(self) -> None:
        await mount_about_header_icon(
            current_app=self,
            icon=ICON,
            app_metadata=APP_METADATA,
        )
        self.translate_about_header_icon()

        self.create_new_game()

    def translate_bindings(self) -> None:
        for key, binding in KEY_TO_BINDING.items():
            current_binding: Binding = self._bindings.key_to_bindings[key][0]
            self._bindings.key_to_bindings[key] = [
                dataclasses.replace(current_binding, description=_(binding.description))
            ]

    def translate_about_header_icon(self) -> None:
        about_header_icon: AboutHeaderIcon = self.query_one(AboutHeaderIcon)
        about_header_icon.tooltip = _("About")

    def translate(self) -> None:
        self.translate_bindings()
        self.translate_about_header_icon()

    def create_new_game(self):
        if hasattr(self, "game"):
            self.board.remove()

        self.game = Game()

        self.board = Board(self.game)
        self.mount(self.board)

    async def on_click(self, event: Click) -> None:
        if isinstance(event.widget, Widget):
            if event.widget.id == "check":
                await self.on_click_check()

    async def on_click_check(self) -> None:
        code_peg_values: list[int] = []
        for code_peg in self.board.current_row.code_pegs:
            code_peg.query_one("SelectCurrent Static.down-arrow").remove()

            code_peg_value: int
            if isinstance(code_peg.value, int):
                code_peg_value = code_peg.value
            else:
                code_peg_value = 0

            code_peg_values.append(code_peg_value)

        num_red_pegs: int
        num_white_pegs: int
        num_red_pegs, num_white_pegs = await self.game.check_code(
            breaker_code=code_peg_values
        )

        self.board.current_row.query_one("#check").remove()

        self.board.current_row.mount(
            Label(
                "".join(
                    [
                        (FEEDBACK_PEG_COLORS[0] + " ") * num_red_pegs,
                        (FEEDBACK_PEG_COLORS[1] + " ") * num_white_pegs,
                        (BLANK_COLOR + " ")
                        * (self.game.num_pegs - num_red_pegs - num_white_pegs),
                    ]
                ),
                classes="feedback_pegs",
            )
        )

        self.board.current_row.disabled = True

        if num_red_pegs == self.game.num_pegs:
            self.notify(_("Congratulations!"))
        else:
            if self.board.current_row_number < self.game.num_rows:
                self.board.add_row()
            else:
                maker_code: list[int] = self.game.get_maker_code()
                maker_code_str: str = ""
                for color in maker_code:
                    if color == 0:
                        maker_code_str += BLANK_COLOR + " "
                    else:
                        maker_code_str += CODE_PEG_COLORS[color - 1] + " "

                self.notify(
                    f"{_('Better luck next time')}\n{_('Code')}: {maker_code_str}",
                    timeout=30,
                )

    @work
    async def action_new_game(self) -> None:
        if await self.push_screen_wait(
            ConfirmScreen(
                dialog_title="New game",
                dialog_subtitle=APP_METADATA.name,
                question="Are you sure you want to start a new game?",
            )
        ):
            self.create_new_game()

    @work
    async def action_settings(self) -> None:
        setting_rows: dict[str, SettingRow] = {
            key: value
            for key, value in zip(
                app_settings.__dict__.keys(),
                [
                    SettingRow(
                        label="Language:",
                        widget=Select(
                            options=zip(
                                [_(value) for value in LANGUAGES.values()],
                                LANGUAGES.keys(),
                            ),
                            value=app_settings.language,
                            allow_blank=False,
                        ),
                    ),
                    SettingRow(
                        label="Variation:",
                        widget=Select(
                            options=zip(
                                [
                                    variation.description
                                    for variation in VARIATIONS.values()
                                ],
                                VARIATIONS.keys(),
                            ),
                            value=app_settings.variation.name,
                            allow_blank=False,
                        ),
                    ),
                    SettingRow(
                        label="Duplicate colors:",
                        widget=Switch(value=app_settings.duplicate_colors),
                    ),
                    SettingRow(
                        label="Blank color:",
                        widget=Switch(value=app_settings.blank_color),
                    ),
                ],
            )
        }

        settings_dict = await self.push_screen_wait(
            SettingsScreen(
                dialog_title="Settings",
                dialog_subtitle=APP_METADATA.name,
                setting_rows=setting_rows,
            )
        )

        if settings_dict is not None:
            old_language = app_settings.language

            old_variation = app_settings.variation
            old_duplicate_colors = app_settings.duplicate_colors
            old_blank_color = app_settings.blank_color

            app_settings.set(settings_dict)

            if old_language != app_settings.language:
                set_translation(app_settings.language)
                self.translate()

            if (
                old_variation.name != app_settings.variation.name
                or old_duplicate_colors != app_settings.duplicate_colors
                or old_blank_color != app_settings.blank_color
            ):
                self.notify(
                    _("New game settings will be applied to a new game"), timeout=5
                )

            save_settings(settings_dict, SETTINGS_PATH)



import random

from mastermind.settings import app_settings


class Game:
    def __init__(self) -> None:
        self.num_rows = app_settings.variation.num_rows
        self.num_pegs = app_settings.variation.num_pegs
        self.num_colors = app_settings.variation.num_colors

        colors: list[int] = list(range(1, self.num_colors + 1))
        if app_settings.blank_color:
            colors.append(0)

        self.maker_code: list[int]
        if not app_settings.duplicate_colors:
            self.maker_code = random.sample(colors, k=self.num_pegs)
        else:
            self.maker_code = random.choices(colors, k=self.num_pegs)

    async def check_code(self, breaker_code: list[int]) -> tuple[int, int]:
        num_red_pegs: int = 0
        num_white_pegs: int = 0

        red_idxs: list[int] = []
        for i, breaker_code_color in enumerate(breaker_code):
            if self.maker_code[i] == breaker_code_color:
                num_red_pegs += 1
                red_idxs.append(i)

        breaker_code_no_reds = [
            color for i, color in enumerate(breaker_code) if i not in red_idxs
        ]
        maker_code_no_reds = [
            color for i, color in enumerate(self.maker_code) if i not in red_idxs
        ]

        for color in breaker_code_no_reds:
            if color in maker_code_no_reds:
                num_white_pegs += 1
                maker_code_no_reds.remove(color)

        return num_red_pegs, num_white_pegs

    def get_maker_code(self) -> list[int]:
        return self.maker_code



from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Label

from mastermind.game import Game
from mastermind.widgets.check import Check
from mastermind.widgets.code_peg import CodePeg


class Row(Horizontal):
    def __init__(self, game: Game, row_number: int) -> None:
        super().__init__(classes="row")

        self.game = game

        self.row_number = row_number

        self.code_pegs: list[CodePeg] = [
            CodePeg(self.game) for _ in range(self.game.num_pegs)
        ]

        self.check: Check = Check(self.game)

    def compose(self) -> ComposeResult:
        yield Label(f"{self.row_number:02}", classes="num")
        for code_peg in self.code_pegs:
            yield code_peg
        yield self.check
